/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _expressions_h_
#define _expressions_h_

#include "Basic.h"
#include "Strings.h"
#include "Defs.h"
#include "Tags.h"
#include "Globals.h"

// class forward references...

class SymbolList ;
class SourceListing ;

/*
 * Descriptors for unparsing C and C++ operator expressions
 */

typedef struct oprec {
  char *opname ;
  int  precedence ;
} oprec_t ;

/*
 * Values, constants, and other non-operator expressions, are all
 * given the highest pseudo-precedence denoted by HYPER_PREC
 */
  
#define HYPER_PREC 20
#define POSTFIX_PREC 16
#define PREFIX_UNARY 15

	/////////////////
	//
	// Every expression is either an LVALUE or an RVALUE.
	// This attribute is indicated by a field of the 
	// enumerated type defined as evalue_t
	//
typedef enum { RVALUE=0, LVALUE=1 } evalue_t ;

/*--------------------------------------------------------------

CLASS
    Expression

    Parse tree object which represents a single expression

DESCRIPTION

 This class implements nodes representing single expressions
 within a parse tree. It is derived publicly from the Tag
 class - thus tagging each expression with the tag of the
 operator or function category it represents.

 Each Expression object contains pointers to the elements of an
 expression, organised as a group of type unions.

----------------------------------------------------------------*/

class Expression : public Tag, private parse_constants {

friend class Statement ;
friend class Symbol ;
friend class FortranContext ;
friend class FortranParser ;
friend class ParseContext ;
friend class Universe<Expression> ;

private:

  int           id ;       // unique identification tag
  PTR_CNTXT     pc ;       // ptr to containing context

// protected: // this will eventually be protected

public:

  PTR_TYPE	type ;	   // expression type
  evalue_t      lvalue ;   // 1 -> lvalue, 0 -> rvalue
  PTR_TRANS     trans ;    // pointer to transformation record
  
  union {
    
    string     *string_val;// for float, double and string constants
    int	        ival ;     // for integer constants
    long        lval ;     // for long integer constants
    unsigned long uval ;   // for unsigned integer constants
    double	dval ;	   // for double-precision constants
    float       fval ;     // for floating point constants
    char	cval ;     // for character constants
    bool	bval;	   // for boolean constants
    
    struct {		   // for range, upper, and lower
      PTR_SYMB  symbol ;
      int	  dim ;
    }			       array_op ;
    
    struct {               // default struct: symbol + 2 sub-expressions
      PTR_SYMB  symbol ;
      PTR_EXPR  e1 ;
      PTR_EXPR  e2 ;
    }			      fields ;
    
    struct {	           // for complex and double complex
      PTR_SYMB  null ;
      PTR_EXPR  real_part ;
      PTR_EXPR  imag_part ;
    }			      complex ;
    
    struct {               // for lists of labels
      PTR_FTN_LABEL lab_ptr ;
      PTR_EXPR  null_1 ;
      PTR_EXPR  next ;
    }			      label_list ;
    
    struct {               // for generic expression lists
      PTR_SYMB  null_1 ;
      PTR_EXPR  item ; 
      PTR_EXPR  next ;
    }			      list ;
      
    struct {               // for cons list
      PTR_SYMB  null_1 ;
      PTR_EXPR  size ;
      PTR_EXPR  list ;
    }			      cons ;
    
    struct {
      PTR_SYMB  control_var ;
      PTR_EXPR  array ;
      PTR_EXPR  range ;
    }			      access ;
    
    struct {
      PTR_SYMB  control_var ;
      PTR_EXPR  array ;
      PTR_EXPR  range ;
    }			      ioaccess ;
    
    struct {
      PTR_SYMB  symbol ;
      PTR_EXPR  null_1 ;
      PTR_EXPR  null_2 ;
    }			      const_ref ;
    
    struct {
      PTR_SYMB  symbol ;
      PTR_EXPR  null_1 ;
      PTR_EXPR  null_2 ;
    }			      var_ref;
    
    struct {
      PTR_SYMB  symbol ;
      PTR_EXPR  index ;
      PTR_EXPR  array_elt ;
    }			      array_ref;

    struct {
      PTR_SYMB  null_1 ;
      PTR_EXPR  access ;
      PTR_EXPR  index ;
    }			      access_ref;
    
    struct {
      PTR_SYMB  null_1 ;
      PTR_EXPR  cons ;
      PTR_EXPR  index ;
    }			      cons_ref;
    
    struct {
      PTR_SYMB  symbol ;
      PTR_EXPR  null_1 ;
      PTR_EXPR  rec_field ;   // for record fields
    }			      record_ref; 
    
    
    struct {
      PTR_SYMB  symbol ;
      PTR_EXPR  param_list ;
      PTR_EXPR  next_call ;
    }			      proc;
   
    struct {
      PTR_SYMB  null_1 ;
      PTR_EXPR  operand ;  
      PTR_EXPR  null_2 ;
    }			      unary_op; 
    
    struct {
      PTR_SYMB  null_1 ;
      PTR_EXPR  l_operand ;
      PTR_EXPR  r_operand ;
    }			      binary_op;
    
    struct {
      PTR_SYMB  null_1 ;
      PTR_EXPR  ddot ;
      PTR_EXPR  stride ;
    }			      seq;
    
    struct {
      PTR_SYMB  null_1 ;
      PTR_EXPR  sp_label ;
      PTR_EXPR  sp_value ;
    }			      spec_pair;
    

    struct {                 // Tag == DECLARATION
      PTR_TYPE    decl_specs ;
      SymbolList* declarator_list ;
      PTR_EXPR    null_2 ;
    }			      declaration;
    
    struct {                  // Tag == INIT_DECLARATOR
      PTR_SYMB  declarator ;
      PTR_EXPR  init_expr ;
      PTR_EXPR  next ;
    }			      declarator_list;
    
    struct {                  // Tag == IMPL_CONV
      PTR_SYMB  null_sym ;
      PTR_EXPR  expr ;
      void     *conv ;
    }			      conversion;
    
    struct {                  // Tag == TYPE_NODE
      PTR_TYPE  dest_type ;
      declaration_tag dtag ;
      PTR_EXPR  null_2 ;
    }			      type_expr;
    
  } node ;

public:

    ////////////////////////////////////
    //
  int index ()                { return id ; }
  // variant provided by Tag class

    ////////////////////////////////////
    //
  PTR_CNTXT context ()        { return pc ; }

    ////////////////////////////////////
    //
  void set_node (PTR_CNTXT c){
    pc = c ;
    type = 0 ;
    trans = 0 ;
  }

    ////////////////////////////////////
    //
  void setType (PTR_TYPE t)   { type = t ; }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, expression_tag v, PTR_EXPR e1, 
		   PTR_EXPR e2, PTR_SYMB sym)
  {
    set_node (c) ;
    expr_variant (v) ;
    node.fields.symbol = sym ;
    node.fields.e1 = e1 ;
    node.fields.e2 = e2 ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, expression_tag v, PTR_FTN_LABEL l) {
    set_node (c) ;
    expr_variant (v) ;
    node.label_list.lab_ptr = l ;
    node.label_list.null_1 = 0 ;
    node.label_list.next = 0 ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, string *s) {
    set_node (c) ;
    expr_variant (STRING_VAL) ;
    node.string_val = s ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, const char *s) {
    set_node (c) ;
    expr_variant (STRING_VAL) ;
    node.string_val = new string (s) ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, int i) {
    set_node (c) ;
    expr_variant (INT_VAL) ;
    node.ival = i ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, long l) {
    set_node (c) ;
    expr_variant (INT_VAL) ;
    node.lval = l ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, unsigned long u) {
    set_node (c) ;
    expr_variant (INT_VAL) ;
    node.uval = u ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, bool b) {
    set_node (c) ;
    expr_variant (BOOL_VAL) ;
    node.bval = b ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, float f) {
    set_node (c) ;
    expr_variant (FLOAT_VAL) ;
    node.fval = f ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT c, double d) {
    set_node (c) ;
    expr_variant (DOUBLE_VAL) ;
    node.dval = d ;
  }

    ////////////////////////////////////
    //
  void set_fields (PTR_CNTXT p, char c) {
    set_node (p) ;
    expr_variant (CHAR_VAL) ;
    node.cval = c ;
  }

    ////////////////////////////////////
    //
  void set_fields (Expression& e) ;
  
    ////////////////////////////////////
    //
    // Actionless constructor
    //
  Expression () : Tag ()
  {
    type = 0 ;
    lvalue = RVALUE ;
    trans = 0 ;
    node.uval = 0 ;
    node.fields.e1 = node.fields.e2 = 0 ;
    node.fields.symbol = 0 ;
  }

    ////////////////////////////////////
    //
    // Normal constructor called by make_member()
    //
  Expression (int i) : Tag (0) { id = i ; }

    ////////////////////////////////////
    //
  ~Expression ()
  {
    switch (expr_variant())
    {
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case STRING_VAL:
    case STMT_STR:
    case KEYWORD_VAL:
      delete node.string_val ;
      break ;
    default:
      break ;
    }
  }

    ////////////////////////////////////
    //
  void set_expr_type () ;
  int is_array_section_ref () ;
  int is_substring_ref () ;
  int match_an_arg (PTR_SYMB) ;
  type_tag typing () ;
  type_tag base_type () ;

    ////////////////////////////////////
    //
    // Returns true if this Expression can be statically evaluated to
    // an integer constant. If possible, the integer pointed to by the
    // int* argument is given the actual integer value of this
    // Expression.
    //
    // [in] int* pointer to result integer
    // [out] bool true if Expression is a constant integer
    //
  bool isIntConst (int *) ;

    ////////////////////////////////////
    //
    // Checks whether this expression can be cast implicitly
    // to the target type given by the second argument.
    // This function may generate error messages via
    // calls to the ParseContext::err() function associated
    // with the context argument.
    // The action causing this attempted coercion is given
    // as a character string in the second argument. It may
    // be "assign", "return", or initialise".
    //
    // [in] Datatype* pointer to target type
    // [in] char* pointer to action
    // [out] char* pointer to created error string or null if no error
    //
  char* coerceToType (PTR_TYPE, char*) ;

    ////////////////////////////////////
    //
  PTR_EXPR tail () {
    PTR_EXPR t = this ;
    while (t && t->node.fields.e2)
      t = t->node.fields.e2 ;
    return t ;
  }

    ////////////////////////////////////
    //
  void append (PTR_EXPR e) {
    if (!e) return ;
    PTR_EXPR temp = this->tail () ;
    temp->node.list.next = e ;
  }

    ////////////////////////////////////
    //
  PTR_FTN_LABEL label () { return node.label_list.lab_ptr ; }

    ////////////////////////////////////
    //
    // Annotates this Expression with a pointer to a transformation
    // record describing the transformation applied to this node.
    //
  void note_transform (PTR_TRANS t) { trans = t ; }

    ////////////////////////////////////
    //
    // Returns a pointer to a TransHist object describing
    // describing the transformation applied to this node.
    // May return a null pointer if there was no transformation.
    //
  PTR_TRANS transformation ()       { return trans ; }

    ////////////////////////////////////
    //
    // Given a pointer to an Expression, this function
    // returns a pointer to the symbol if this expression 
    // is either an ARRAY_REF or a VAR_REF node. Zero is
    // returned if this Expression is of any other kind.
    //
    //  [out] Symbol*     pointer to the referenced symbol
    //
  PTR_SYMB referenced_symbol ()
  {
    if (expr_variant() == ARRAY_REF || expr_variant() == VAR_REF)
      return node.fields.symbol ;
    else
      return 0 ;
  }

    ////////////////////////////////////
    //
  int exprListLength ()
  {
    int len = 0 ;
    PTR_EXPR l = this ;
    while (l)
      {
	++len ;
	l = l->node.fields.e2 ;
      }
    return len ;
  }

    ////////////////////////////////////
    //
  PTR_EXPR exprListItem (int i)
  {
    PTR_EXPR n = this ;
    while (n && i)
      {
	n = n->node.fields.e2 ;
	--i ;
      }
    if (n)
      return n->node.fields.e1 ;
    else
      return 0 ;
  }

    ////////////////////////////////////
    //
    // Returns true if this node has a symbol, otherwise
    // it returns false.
    //
  bool hasNodeASymbol () ;

    ////////////////////////////////////
    //
  void replaceSymbol (PTR_SYMB os, PTR_SYMB ns)
    {
      if (!this)
	return ;

      if (hasNodeASymbol())
	if (node.fields.symbol == os)
	  node.fields.symbol = ns ;
      node.fields.e1->replaceSymbol (os, ns) ;
      node.fields.e2->replaceSymbol (os, ns) ;
    }

    ////////////////////////////////////
    //
  void CcUnparseExpression (SourceListing&, int) ;

    ////////////////////////////////////
    //
  void FtnUnparseExpression (LineBuffer&) ;
};

#endif // _expressions_h_

