/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// Header for Parser Context class - the base-level class which
// defines the context within which all parsing takes place, 
// irrespective of source language.

#if ! defined _parseContext_h_
#define _parseContext_h_

#include "Sets.h"
#include "Strings.h"
#include "Globals.h"
#include "ParseTree.h"
#include "TransHistory.h"

// class forward references

class ParseContext ;

// constant definitions

#define TBUF_SIZE 1024

// CLASS GROUP: parser support classes

/*--------------------------------------------------------------------

CLASS
    SourceFile

    Object for recording details of a single source file.

OVERVIEW

    The SourceFile class is used for recording the details of
    a single source file within the include-tree of some
    compilation unit.

DESCRIPTION

    A SourceFile is an enumerable string, which also contains
    a list of SourceFile object indices representing the files included
    from this SourceFile (in the order in which they are included).
    Each SourceFile also poitns to the file which
    included this SourceFile, to permit a trace-back from the
    current file to the root source file.

    Each ParseContext object contains a Universe of SourceFile
    objects, from which any individual SourceFile can be extracted
    if its index is known. It also contains a pointer back to its
    containing ParseContext.

    The ParseContext must call
    'new SourceFile (filename, index, parent, parsecontext)'
    in order to create a new SourceFile object at the point of inclusion.
    The index is maintained by the ParseContext by keeping a count of the
    number of included files in the compilation unit, which is then
    incremented after creation of each new SourceFile object.

    A negative value for the parent index indicates that the file
    has no parent, i.e. it is the root file. This is also indicated
    by a null parent pointer in the root SourceFile object.

---------------------------------------------------------------------*/

class SourceFile : public Enumerable {

public:

  SourceFile (const char* fn, int ix, SourceFile *par, ParseContext *cntxt)
    : Enumerable (ix)
  {
    fname = new string (fn) ;
    p = par ;
    context = cntxt ;
    incfiles = 0 ;
    if (p)
      p->appendIncludeFile (this) ;
  }

  ~SourceFile () { }

  bool isRootFile () { return (bool)!p ; }

  int parentIndex () { return p ? p->index() : -1 ; }

  SourceFile *parentFile () { return this ? p : 0 ; }

  const char* fileName () { return (this && fname) ? fname->operator()() : 0 ; }

  void appendIncludeFile (SourceFile *incf)
  {
    if (!incfiles)
      incfiles = new dlist<int,SourceFile> () ;
    incfiles->append (incf) ;
  }
  
  char *include_trace ()
  {
    char buff [TBUF_SIZE] ;
    char *bp ;
    SourceFile *ip ;
    int tl ;
    int sl ;
    buff[0] = '\0' ; /* just in case trace is empty */

    for (ip = p, tl = 0, bp = buff ; ip ; )
      {
	sprintf (bp, "\tin file included from: '%s'\n\000", 
		 ip->fileName() ) ;
	bp += (sl = strlen (bp)) ;
	tl += sl ;
	if (tl > (TBUF_SIZE - 100))
	  {
	    ip = 0 ;
	    sprintf (bp, "\t...too many include files...\n") ;
	  }
	else
	  ip = ip->p ;
      }
    tl = strlen (buff) + 1 ;
    char *r = new char [tl] ;
    strcpy (r, buff) ;
    return r ;
  }
  
  int variant() { return 0 ; }

private:

  string *fname ;
  SourceFile *p ;
  ParseContext *context ;
  dlist<int,SourceFile> *incfiles ;
};


/*--------------------------------------------------------------------

CLASS
    Keylist

    Tuple for holding one entry in a list of keywords

OVERVIEW
 
    Each language front end has a list of reserved keywords, declared
    as a statically-defined array of Keylist tuples. These contain a
    const char* pointer to the actual reserver word and a const int
    key value.

------------------------------------------------------------------------*/

struct Keylist {
public:
  char *keyname ;
  int  keyval ;
};

        ////////////////////////////
        // Enumerated list of common lexical errors, used by all front ends.
        //
typedef enum lexErrorCodes {
  UNKNOWN_ERROR = -1,
  OCT_RANGE = -2,
  HEX_RANGE = -3,
  SEPARATOR_IN_STRING = -4,
  NOT_SINGLE_CHAR = -5,
  UINT_RANGE = -6,
  ULINT_RANGE = -7,
  ICONST_RANGE = -8,
  LICONST_RANGE = -9,
  SILENT_ERROR = -10,
  RADIX_ERROR = -11,
  LEX_ERROR_LIMIT = -12,
} lexErrorTag;


/*----------------------------------------------------------
CLASS
     Token

     A template class which holds one complete lexical token.

OVERVIEW

     A lexical token consists of an integer token code, and
     possibly also a token value. The value is a single
     word entity of type T, the argument type to this template
     class. We construct a tuple 
     containing both of these items when a Token is
     created. Certain token values are pointers to other
     objects. These other objects are not owned by the
     Token object, but are simply refered to.

 ----------------------------------------------------------*/

template <class T> class Token {
public:
  int tcode ;
  T tval ;
  
  Token () : tval(0) 
  { 
    tcode = 0 ;
  }

  Token (int t, T v) : tval(v)
  { 
    tcode = t;
  }

  Token (int t, T* v = 0) : tval(v ? v->uval : 0)
  {
    tcode = t;
  }
  
  int index() { return 0 ; }        // should never be needed
  int variant() { return tcode ; }  // should never be needed
};
 

/*----------------------------------------------------------
CLASS
     TokenQueue

     A class which implements a queue of tokens. 

OVERVIEW
 
     Each language recogniser has a lexical analyser which
     may require to use a queue of tokens. The tokens have
     a lexical value of type T, the argument to this
     template type. Typically each lexers header file
     will create a unique specialisation of this class.

     At certain points during parsing it is necessary to
     be able to scan ahead in the sequence of tokens in
     order to disambiguate parts of the grammar. 
     This is achieved by buffering the tokens in a
     TokenQueue and subsequently directing the lexer()
     function to return tokens from that queue before it
     scans any new tokens.

     Deferred parsing can be implemented by scanning a
     sequence of tokens into a TokenQueue and later on
     enabling parser input from that queue.

 ----------------------------------------------------------*/

template <class T> class TokenQueue : public slist<int,Token<T> > {
public:
  
  TokenQueue () { has_tokens = false ; }

  void enqueue (int tok, T lval) { append (new Token<T>(tok, lval)); }

  void prepend (int tok, T lval) { slist<int,Token<T> >::prepend (new Token<T>(tok, lval)); }

  int dequeue (T* lvalp)
  {
    Token<T> *ntok = get() ;
    if (ntok)
      {
	int code = ntok->tcode ;
	lvalp->uval = ntok->tval.uval ;
	delete ntok ;
	return code ;
      }
    else
      {
	lvalp->uval = 0 ;
	has_tokens = false ;
	return 0 ;
      }
  }

        ///////////////////
        //
        // Initialises an internal iterator
        //
  void beginScan ()
  {
    next_token = this ;
    has_tokens = true ;
  }

        ///////////////////
        //
        // Returns next enqueued token without
        // modifying the queue. Returns -1 if
        // there are no more buffered lookahead
        // tokens in the queue.
        //
  int next_lookahead (T* lvalp)
  {
    Token<T> *t = has_tokens ? next_token() : 0 ;
    if (!t)
      {
	has_tokens = false ;
	return -1 ;
      }
    else
      {
	lvalp->uval = t->tval.uval ;
	return t->tcode ;
      }
  }

  int has_any_buffered_tokens () { return has_tokens ; }

  int index ()  { return 0 ; } // these two methods should never
  int variant() { return 0 ; } // actually be needed (keeps g++ happy).

private:
  
  slist_iterator<int,Token<T> > next_token ;
  int has_tokens ;
};


/*--------------------------------------------------------------------
  
CLASS
    ParseContext

    A general context for parsing source code.

OVERVIEW

    You should only be concerned about ParseContext classes if you
    are considering creating a new parser within the Octave library.

    If you are only interested in using existing languages, then
    all details of the ParseContext are hidden underneath the
    language-specific parser classes, such as FortranParser.

DESCRIPTION

    The ParseContext class is provided as a base class from which
    specific language parser classes can be derived.

    The ParseContext class has two purposes. Firstly as a container
    for all the parse-tree entities created during parsing, and
    secondly as a provider of member functions for constructing
    new parse tree entities.

    Every parse tree object, whether it is a Statement, an Expression,
    a Symbol, a Label, or a Datatype, exists as an element within a
    Universe of those objects. As such, one can create a SetOf each
    class of parse tree object. These universes are private data
    members within each ParseContext.

    Each ParseContext is associated with a single source file,
    and has a root statement representing the root of the parse
    tree for that file. It also maintains a list of parse errors
    encountered during parsing.

    Functions are provided for error handling during parsing, so
    any language-specific parser class simply inherits these
    functions and a consistent error handling interface is
    presented to the user.

    Functions are provided for the creation of hash table entries,
    symbols, expressions, statements, labels, types and comments.
    
---------------------------------------------------------------------*/

class ParseContext : public parse_constants,
                     public ScopeStack       {

friend class FortranParser ;
friend class CcParser ;
friend class FortranContext ;
friend class CcContext ;
public:

  // GROUP: publicly available data members

           //////////////////////////
           //
           // Set this to any non-zero value to
           // enable generation of warning messages
           // during compilation.
           //
  int      nowarnflag ;

           //////////////////////////
           //
           // During parsing, this variable contains the
           // source line number at which the current
	   // statement begins.
           //
  int      yylineno ;

           //////////////////////////
           //
           // During parsing, this variable contains the
           // source line offset from the start of a
           // module.
           //
  int      mod_offset ;

  // GROUP: predicates and identifying functions


           //////////////////////////
           //
  int index () { return id ; }

           //////////////////////////
           //
           // Returns the language which this ParseContext
           // is defined to contain.
           //
  language OctLanguage ()  { return lang ; }

  // GROUP: creation of atomic parse-tree entities


           //////////////////////////
           //
  PTR_HASH make_hash_entry (const char *string) ;

           //////////////////////////
           //
  PTR_SYMB make_symb (symbol_tag, const char *) ;

           //////////////////////////
           //
  PTR_SYMB make_symb (Symbol&) ;

           //////////////////////////
           //
  PTR_SYMB make_symb (symbol_tag, string *) ;

           //////////////////////////
           //
  PTR_SYMB make_symb (symbol_tag, PTR_HASH, int = 0) ;

           //////////////////////////
           //
  PTR_SYMB make_symb () ;

           //////////////////////////
           //
  PTR_EXPR make_expr (expression_tag, PTR_EXPR, PTR_EXPR, PTR_SYMB = 0) ;

           //////////////////////////
           //
           // The following  version of make_expr is used extensively by the C++ front
           // end. It allows the LVALUE/RVALUE attribute to be defined, as
           // well as the type of the expression. The left and right sub-expressions
           // are specified as the final two fields.
           //
  PTR_EXPR make_expr (expression_tag, evalue_t, PTR_TYPE, PTR_EXPR, PTR_EXPR=0) ;

           //////////////////////////
           //
  PTR_EXPR make_expr () ;

           //////////////////////////
           //
           // Creates a constant long integer expression
           //
  PTR_EXPR make_int_const (PTR_TYPE, long) ;

           //////////////////////////
           //
           // Useful for creating a constant character expression
           //
  PTR_EXPR make_char_const (PTR_TYPE, char) ;

           //////////////////////////
           //
           // Useful for creating a wide character expression
           //
//  PTR_EXPR make_wchar_const (PTR_TYPE, wchar_t) ;

           //////////////////////////
           //
           // Useful for creating a string (or char*) constant expression
           //
  PTR_EXPR make_string_const (PTR_TYPE, string*) ;

           //////////////////////////
           //
           // Useful for creating a floating-point constant stored as a string
           //
  PTR_EXPR make_float_const (PTR_TYPE, string*) ;


           //////////////////////////
           //
           // Useful for creating a string (or char*) constant expression
           //
  PTR_EXPR make_bool_const (PTR_TYPE, bool) ;

           //////////////////////////
           //
           //
  PTR_EXPR make_expr (Expression&) ;

           //////////////////////////
           //
  PTR_EXPR make_expr (expression_tag, PTR_FTN_LABEL) ;

           //////////////////////////
           //
  PTR_EXPR make_expr (const char *) ;

           //////////////////////////
           //
  PTR_EXPR make_expr (string *) ;

           //////////////////////////
           //
  PTR_EXPR make_expr (int) ;

           //////////////////////////
           //
  PTR_EXPR make_expr (bool) ;

           //////////////////////////
           //
  PTR_EXPR make_expr (float) ;

           //////////////////////////
           //
  PTR_EXPR make_expr (double) ;

           //////////////////////////
           //
  PTR_EXPR make_expr (char) ;

           //////////////////////////
           //
  PTR_EXPR set_ll_list (PTR_EXPR, PTR_EXPR, expression_tag) ;

           //////////////////////////
           //
  PTR_EXPR append_elist (PTR_EXPR, PTR_EXPR);

           //////////////////////////
           //
  PTR_STMT make_stmt (Statement&) ;

           //////////////////////////
           //
  PTR_STMT make_stmt (statement_tag, PTR_SYMB, PTR_EXPR, PTR_EXPR, PTR_EXPR) ;

           //////////////////////////
           //
  PTR_STMT make_stmt (statement_tag, PTR_SYMB, PTR_EXPR, PTR_EXPR, PTR_EXPR,
		      int, int) ;

           //////////////////////////
           //
  PTR_STMT make_stmt () ;

           //////////////////////////
           //
  PTR_FTN_LABEL make_label (int) ;

           //////////////////////////
           //
  PTR_TYPE make_type (type_tag, int=0) ;

           //////////////////////////
           //
  PTR_TYPE make_type_spec (int) ;

           //////////////////////////
           //
  PTR_TYPE make_type_node (PTR_TYPE, PTR_EXPR) ;

           //////////////////////////
           //
  PTR_TYPE install_array (PTR_EXPR, PTR_TYPE, int) ;

           //////////////////////////
           //
  PTR_CMNT make_comment (const char *, int) ;

           //////////////////////////
           //
  PTR_CMNT make_comment (string *, int) ;

  // GROUP: creation of transformation records

           //////////////////////////
           //
  PTR_TRANS make_transform (trans_tag, PTR_STMT) ;

           //////////////////////////
           //
  PTR_TRANS make_transform (trans_tag, PTR_SYMB) ;

           //////////////////////////
           //
  PTR_TRANS make_transform (trans_tag, PTR_EXPR) ;

           //////////////////////////
           //
  PTR_TRANS make_transform (trans_tag, PTR_TYPE) ;

  // GROUP: error routines

           //////////////////////////
           //
  void err (const char *) ;

           //////////////////////////
           //
  void error (const char *) ;

           //////////////////////////
           //
  void errstr (const char *, const char *) ;

           //////////////////////////
           //
  void erri (const char *, int) ;

           //////////////////////////
           //
  void warn (const char *) ;

           //////////////////////////
           //
  void warni (const char *, int) ;

           //////////////////////////
           //
  void warnstr (const char *, char *) ;

           //////////////////////////
           //
  void fatal (const char *) ;

           //////////////////////////
           //
  void fatalstr (const char *, const char *) ;

           //////////////////////////
           //
  void fatali (const char *, int) ;

           //////////////////////////
           //
  int  nerrors () { return nb_errors ; }
  
  // GROUP: construction and destruction


           //////////////////////////
           //
           // Constructor.
           //
           //  [in]  string    filename for this context
           //  [in]  language  the input language
           //
           //
  ParseContext (string&, language) ;
  
           //////////////////////////
           //
           // Constructor.
           //
           //  [in]  char*     filename for this context
           //  [in]  language  the input language
           //
           //
  ParseContext (const char*, language) ;
  

           //////////////////////////
           // 
           // Destructor.
           //
  ~ParseContext() ;


           //////////////////////////
           //
           // Return the PTR_STMT which is the root of
           // the parse tree for this parse context.
           //
  PTR_STMT fileRoot() { return root_stmt ; }

           //////////////////////////
           //
           // Return the PTR_SYMB which is the head of
           // chain of all Symbols in this context.
           //
  PTR_SYMB first_symb() { return root_symb ; }


           //////////////////////////
           //
           // Return a reference to the the filename
           // associated with this ParseContext.
           //
  string& filename ()   { return fname ; }

  // GROUP: locating parse-tree entities by ID

           //////////////////////////
           //
  PTR_STMT find_statement (int i)  { return statement_domain->get_member (i) ; }

           //////////////////////////
           //
  PTR_EXPR find_expression (int i) { return expression_domain->get_member (i) ; }

           //////////////////////////
           //
  PTR_SYMB find_symbol (int i)     { return symbol_domain->get_member (i) ; }

           //////////////////////////
           //
  PTR_TYPE find_type (int i)       { return type_domain->get_member (i) ; }

           //////////////////////////
           //
  PTR_CMNT find_comment (int i)    { return comment_domain->get_member (i) ; }

           //////////////////////////
           //
  PTR_TRANS find_transform (int i)  { return trans_domain->get_member (i) ; }

  // GROUP: retrieval of AST object universes

           //////////////////////////
           //
  Universe<Statement>* theStatements () { return statement_domain ; }

           //////////////////////////
           //
  Universe<Expression>* theExpressions () { return expression_domain ; }

           //////////////////////////
           //
  Universe<Symbol>* theSymbols () { return symbol_domain ; }

           //////////////////////////
           //
  Universe<Datatype>* theTypes () { return type_domain ; }

           //////////////////////////
           //
  Universe<Comment>* theComments () { return comment_domain ; }

  // GROUP: error handling functions

           //////////////////////////
           //
           // Returns a singly-linked list of error and
           // warning messages produced during parsing.
           //
  slist<char*,string>& errors ()   { return err_list ; }

  // GROUP: file management functions

           //////////////////////////
           //
           // Given a SourceFile index, lookup the corresponding
           // SourceFile object and return a pointer to it (or null).
           //
  SourceFile *lookupSourceFile (int f)
  {
    return file_list ? file_list->find (f) : 0 ;
  }

           //////////////////////////
           //
           // When parsing, the preprocessor usually indicates
           // that the current file has changed. When this
           // happens, the parser should call setSourceFile, with
           // the new current file as the argument. 
           //
           // This then checks whether the new current file is
           // the outer include file, in which case it switches
           // back to that one. Otherwise it creates a new
           // SourceFile object and links it into the tree of
           // include files.
           //
  void setSourceFile (const char *fn, int f1, int f2)
  {
    if (f1 == 2)
      {
	// We are returning to a previous file
	curSrc = curSrc->parentFile() ;
	return ;
      }
    else if (f1 == 1)
      {
	// We are entering a new include file
	curSrc = new SourceFile (fn, nextSourceFileID++, curSrc, this) ;
	if (!file_list)
	  file_list = new dlist<int,SourceFile> () ;
	
	file_list->append (curSrc) ;
      }
    else if (!f1 && !f2 && !curSrc)
      {
	// We are entering the root file
	curSrc = new SourceFile (fn, nextSourceFileID++, 0, this) ;
	if (!file_list)
	  file_list = new dlist<int,SourceFile> () ;
	
	file_list->append (curSrc) ;
      }
  }

  ObjectMap<Symbol,SetOf<Expression> > *sym_uses ;

private:

  void link_symb (PTR_SYMB) ;
  void link_stmt (PTR_STMT) ;

  int                  id ;
  language             lang ;
  string               fname ;
  string               path ;
  int                  nb_errors ;
  PTR_STMT             root_stmt ;
  PTR_STMT             stmt_head ;
  PTR_STMT             last_stmt ;
  PTR_SYMB             root_symb ;
  PTR_SYMB             last_symb ;
  slist<char*,string>  err_list ;
  Universe<Statement> *statement_domain ;
  Universe<Expression>*expression_domain ;
  Universe<Symbol>    *symbol_domain ;
  Universe<Datatype>  *type_domain ;
  Universe<Comment>   *comment_domain ;
  Universe<TransHist> *trans_domain ;
  dlist<int,SourceFile>*file_list ;
  int                  nextSourceFileID ;
  SourceFile          *curSrc ;
  static int           nextParseContextID ;
} ;

#endif // _parseContext_h_
