/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _types_h_
#define _types_h_

#include "Defs.h"
#include "Tags.h"
#include "Lists.h"

// Forward references to classes that are mentioned before being defined

class TypeStack ;
class TypeList ;
class TypeListIterator ;
class SymbolStack ;
class SymbolList ;
class SourceListing ;

enum UnpCntl {
  UNP_BRIEF,
  UNP_VERBOSE
};

/*--------------------------------------------------------------------+

CLASS
    Datatype

    An object to represent a type node in the internal representation
    of a program.

DESCRIPTION
    This holds all information relating to a single type node. Built-in
    types require one such node, with the base class Tag identifying
    the actual type (from the set of possible type_tag enumerations).
    
----------------------------------------------------------------------*/

class Datatype : public Tag {

friend class TypeIterator ;
friend class Symbol ;
friend class ParseContext ;
friend class Universe<Datatype> ;

private:  
  int id ;                      // unique integer identifier

public:
  PTR_SYMB	name;		// type name
  PTR_CNTXT     pc ;            // ptr to enclosing parse context
  int           length ;        // size of type in bytes
  int           type_flags ;    // bitvec of modifiers defined Tag.h

  union{
  
    struct {		// for T_SUBRANGE
      PTR_TYPE base_type;	// one of { T_INT, T_CHAR, T_FLOAT }
      PTR_EXPR lower;           // ptr to lwb expression
      PTR_EXPR upper;           // ptr to upb expression
      int lwb, upb;             // lwb/upb values
    } subrange;
    
    struct {		// for T_ARRAY
      PTR_TYPE base_type;       // array of this type
      int num_dimensions;       // number of dimensions so far
      PTR_EXPR ranges;          // ptr to dimension expression
    } ar_decl;
    
    struct { 	                // for T_DESCRIPT, T_FUNCTION, T_POINTER
      PTR_TYPE base_type ;      // return type or type pointed to
      TypeList *arg_types ;     // formal parameter type list
      PTR_EXPR ranges ;         // expression list of formal parameters
      int dummy1 ;              // used to initalize all variants
      int dummy2 ;              // used to initalize all variants
      int dummy3 ;              // used to initalize all variants
    } fields ;

    TypeList *list ;            // for T_LIST (a list of type nodes)
    
    struct {		        // for T_RECORD or T_ENUM
      int num_fields;
      int record_size;
      PTR_SYMB first;
    } re_decl;
				
    PTR_SYMB	 derived_type ; // for type name deriving type
    
    struct {		        // for class T_CLASS T_UNION T_STRUCT
      int num_fields;           // number of members 
      int record_size ;         // size of object of this type
      PTR_SYMB first ;          // id-list of member symbols
      PTR_STMT original_class ; // pointer to the declaring statement
      TypeList *obj_type ;      // type list defining object type
      PTR_BLOCK decls ;         // Block of declarations for unparsing
      PTR_SCOPE members ;       // SymbolTable of object scope
    } derived_class ;
        
    struct {		// for class T_DERIVED_COLLECTION
      PTR_SYMB collection_name;
      PTR_TYPE base_type;
    } col_decl ;
    
    struct {		// probably not needed anymore, but used by C++
      PTR_TYPE base_type ;
      PTR_EXPR ranges    ;
      int access_flag;
    } descriptive ;
    
  } node ;

public:

  Datatype () : Tag (0) { }
  Datatype (int i) : Tag (0) { id = i ; }
  Datatype (Datatype& d) ;
  ~Datatype () ;

  int  index () { return id ; }
  // variant() provided by Tag class


  /////////////////////////////
  //
  void set_fields (PTR_CNTXT, type_tag, int=0) ;

  /////////////////////////////
  //
  PTR_TYPE make_type_node (PTR_EXPR lengspec) ;

  // GROUP: type equivalence functions

  /////////////////////////////
  //
  friend int operator== (Datatype &, Datatype &) ;

  /////////////////////////////
  //
  friend int operator!= (Datatype &, Datatype &) ;

  // GROUP: type conversion functions
  
  /////////////////////////////
  //
  // Performs the usual arithmetic conversions applied to pairs of
  // operands given to symetric dyadic arithmetic operators.
  // The type node pointer arguments should be the addresses of
  // the Datatype* fields within a pair Expression operands. 
  // After a call to this function one of the operand types
  // may have been altered to to reflect a type conversion.
  //
  // If there are any errors during the arithmetic conversions
  // a static error message is returned. Otherwise a null 
  // pointer is returned.
  //
  // When a built-int type is promoted or converted the address
  // of a PTR_TYPE variable is passed as argument. That may get
  // overwritten by the destination type, in which case the
  // original type is stored in the node.subrange.base_type
  // field of the newly-created destination type node.
  //
  friend char* usual_conversions (PTR_TYPE*, PTR_TYPE*) ;

  /////////////////////////////
  //
  // If the given type node is not identical with the desired
  // built-in type (as defined by the type tag and type flags),
  // it is replaced with the defined type provided such a
  // conversion is legal. If it not legal a static error message
  // string is returned.
  //
  friend char* convert_type (PTR_TYPE*, type_tag, unsigned long) ;

  /////////////////////////////
  //
  // A char, a short int, enumerator or an int bit-field
  // is converted to an integer. If the given type is not
  // one of the above, or a straightforward integer, an
  // error message is returned. Otherwise a null
  // pointer is returned.
  //
  friend char* integral_promotion (PTR_TYPE*) ;

  /////////////////////////////
  //
  // The standard set of pointer conversions is applied to the
  // type of the expression argument in order to bring it to a pointer type
  // compatible with the second argument.
  // If it is not possible to bring it to a pointer type
  // an error message is returned. The caller must deallocate the
  // memory allocated to the error message.
  //
  friend char* pointer_conversions (PTR_EXPR, PTR_TYPE) ;

  // GROUP: type predicates

  /////////////////////////////
  //
  bool isArithType ()
  {
    type_tag tt = type_variant() ;
    if (tt == T_DESCRIPT)
      return node.fields.base_type->isArithType() ;
    else
      return (bool)IS_ARITH_TYPE(tt) ;
  }

  /////////////////////////////
  //
  bool isIntegralType ()
  {
    type_tag tt = type_variant() ;
    if (tt == T_DESCRIPT)
      return node.fields.base_type->isIntegralType() ;
    else
      return (bool)(tt == T_INT || tt == T_CHAR) ;
  }

  /////////////////////////////
  //
  bool isAggregateType ()
  {
    type_tag tt = type_variant() ;
    if (tt == T_DESCRIPT)
      return node.fields.base_type->isAggregateType() ;
    else
      return (bool)(tt == T_STRUCT || tt == T_UNION || tt == T_CLASS) ;
  }

  /////////////////////////////
  //
  bool isPointerType (unsigned int tf) { return isBuiltInType (T_POINTER, tf) ; }

  /////////////////////////////
  //
  bool isPointerType () { return isBuiltInType (T_POINTER, 0) ; }

  /////////////////////////////
  //
  bool isArrayType () { return isBuiltInType (T_ARRAY, 0) ; }

  /////////////////////////////
  //
  bool isFuncType () { return isBuiltInType (T_FUNCTION, 0) ; }

  /////////////////////////////
  //
  bool isLongDouble () { return isBuiltInType (T_DOUBLE, GOT_LONG) ; }

  /////////////////////////////
  //
  bool isDouble () { return isBuiltInType (T_DOUBLE, 0) ; }

  /////////////////////////////
  //
  bool isFloat () { return isBuiltInType (T_FLOAT, 0) ; }

  /////////////////////////////
  //
  bool isUnsignedLong () { return isBuiltInType (T_INT, GOT_LONG + GOT_UNSIGNED); }

  /////////////////////////////
  //
  bool isLongInt () { return isBuiltInType (T_INT, GOT_LONG); }

  /////////////////////////////
  //
  bool isUnsignedInt () { return isBuiltInType (T_INT, GOT_UNSIGNED); }

  /////////////////////////////
  //
  bool isLong () { return isBuiltInType (T_INT, GOT_LONG); }

  /////////////////////////////
  //
  bool isUnsigned () { return isBuiltInType (T_INT, GOT_UNSIGNED); }

  /////////////////////////////
  //
  bool isChar () { return isBuiltInType (T_CHAR, 0) ; }

  /////////////////////////////
  //
  bool isUnsignedChar () { return isBuiltInType (T_CHAR, GOT_UNSIGNED) ; }

  /////////////////////////////
  //
  bool isBuiltInType (type_tag tt, unsigned long tf)
  {
    if (type_variant() == T_DESCRIPT)
      return node.fields.base_type->isBuiltInType (tt, tf) ;
    else
      return (bool)((type_variant() == tt) && (!tf || (type_flags & tf))) ;
  }

  // GROUP: unparsing and error message generation functions

  /////////////////////////////
  //
  void CcUnparseType (SourceListing&, UnpCntl=UNP_BRIEF) ;

  /////////////////////////////
  //
  // Unparses the type into a newly-created buffer and returns
  // a pointer to that buffer.
  //
  char *sprintf() ;

  /////////////////////////////
  //
  friend char* type_error (char*, PTR_TYPE, PTR_TYPE) ;

  // GROUP: declarator component extraction

  /////////////////////////////
  //
  PTR_TYPE declspec () ;

  /////////////////////////////
  //
  PTR_TYPE dirdecl () ;

  /////////////////////////////
  //
  void CcUnparsePointer (SourceListing&) ;

private:

  void unp_absdecl (SourceListing&, TypeStack&, SymbolStack&, PTR_SYMB) ;
  void unp_dirdecl (SourceListing&, TypeStack&, SymbolStack&) ;
  void unp_declspec (SourceListing&, UnpCntl) ;
} ;

/*--------------------------------------------------------------------+

CLASS
    TypeList

    An object to represent a list of Datatype entities.

DESCRIPTION
    Sometimes it is necessary to represent an aggregate type in terms
    of a root type node and an ordered list of contained types. This
    is the case for classes, structures and unions, for example. These
    are stored as a TypeList object.
    
----------------------------------------------------------------------*/

class TypeList : public slist<int,Datatype>
{ 
public:

  /////////////////////////////
  //
  friend int operator== (TypeList &, TypeList &) ;

  /////////////////////////////
  //
  friend int operator!= (TypeList &, TypeList &) ;
};

/*--------------------------------------------------------------------+

CLASS
    TypeListIterator

    An object for iterating through a TypeList.

DESCRIPTION
    A TypeListIterator can be constructed from a given TypeList, and
    thereafter the () operator can be used to select the next type
    from the list. When the list is exhausted, a null Datatype pointer
    is returned.
    
----------------------------------------------------------------------*/

class TypeListIterator : public slist_iterator<int,Datatype>
{
public:

  /////////////////////////////
  //
  TypeListIterator (TypeList &t) : slist_iterator<int,Datatype> (t) {}

  /////////////////////////////
  //
  TypeListIterator (TypeList *t) : slist_iterator<int,Datatype> (t) {}
};

class TypeStack : public stack<Datatype> { };

#endif // _types_h_
