/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _ftnLexer_h_
#define _ftnLexer_h_

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include "Octave/FlexLexer.h"
#include "Octave/Base/ParseContext.h"
#include "Octave/Fortran/Tokdefs.h"

#define DFT_BSIZE 65536

// Special marker to delimit Hollerith strings

#define HOLL_DELIM (3)

// States for the pre-scanner

enum scan_state_t { 
  IN_PREFIX     = 0, 
  IN_COMMENT    = 1,
  IN_APOSTROPHE = 2,
  IN_QUOTATION  = 3,
  IN_EXCLUDE    = 4,
  IN_HOLLERITH  = 5,
  IN_STATEMENT  = 6
};

/*----------------------------------------------------------
CLASS
     FortranLexer

     A lexical analyser for Fortran.
----------------------------------------------------------*/

class FortranLexer : public yyFlexLexer {

public:

      ///////////////////////////
      //
      // Constructor.
      //
      //  [in] int*      pointer to yylineno variable
      //  [in] istream*  pointer to source input stream
      //  [in] ostream*  pointer to listing output stream
      //  [in] int       optional lexer buffer size
      //
  FortranLexer (int* yylnumPtr, istream* arg_yyin = 0, 
		ostream* arg_yyout = 0, int sz = DFT_BSIZE) 
  : yyFlexLexer (yylnumPtr, &needkwd, arg_yyin, arg_yyout) 
  {
    rawBuf = new char [sz] ;
    scanBuf = new char [sz] ;
    cmntBuf = new char [sz] ;
    buf_size = sz ;
    LexerInit () ;
  }


      ///////////////////////////
      //
      // Constructor.
      //
      //  [in] istream*  pointer to source input stream
      //  [in] ostream*  pointer to listing output stream
      //  [in] int       optional lexer buffer size
      //
  FortranLexer (string& fname, int* yylnumPtr, int sz = DFT_BSIZE) 
  : yyFlexLexer (yylnumPtr, &needkwd) 
  {
    rawBuf = new char [sz] ;
    scanBuf = new char [sz] ;
    cmntBuf = new char [sz] ;
    buf_size = sz ;
    needkwd = 0 ;
    LexerInit () ;
    if (fname())
      {
	in.open (fname(), ios::in) ;
	if (!in)
	  {
	    // later on we will THROW an exception at this point
	    char err_msg[200] ;
	    sprintf (err_msg, "unable to open file '%s'", fname()) ;
	    LexerError (err_msg) ;
	  }
	else
	  yyin = &in ;
      }
  }

      //////////////////////////
      //
      // Destructor
      //
  ~FortranLexer () { delete [] rawBuf ; delete scanBuf ; }

  // GROUP: public data interface from lexer to parser


      //////////////////////////
      //
  int      yylisting ;

      //////////////////////////
      //
  int      yylonglines ;

      //////////////////////////
      //
  long     yystno ;

      //////////////////////////
      //
  int      needkwd ;

      //////////////////////////
      //
  int      colon_flag ;

      //////////////////////////
      //
  char     intonly ;


  // GROUP: public member functions for parser

      ///////////////////////////
      //
      // Lexical analyser used by Fortran parser.
      //
  int lexer (YYSTYPE *lvalp) { 
    int token = yylex(lvalp) ;
    switch (token)
      {
      case COMMENT:
      case SHORT_COMMENT:
      case NAMEEQ:
      case IDENTIFIER:
      case CHAR_CONSTANT:
      case REAL_CONSTANT:
      case HEX_CONSTANT:
      case OCTAL_CONSTANT:
      case DP_CONSTANT:
      case DEFINED_OPERATOR:
      case CONSTRUCT_ID:
	(*lvalp).stringp = new string (YYText()) ;
	break;
      case INT_CONSTANT:
      case LABEL:
	(*lvalp).ival = atoi (YYText()) ;
	break ;
      case HOLLERITH:
	{
	  int l = YYLeng() ;
	  char *hbuf = new char [l + 1] ;
	  strcpy (hbuf, YYText()) ;
	  hbuf[0] = '\'' ;
	  hbuf[l-1] = '\'' ;
	  hbuf[l] = '\0' ;
	  (*lvalp).stringp = new string (hbuf) ;
	  delete [] hbuf ;
	}
	break;
      default:
	break ;
      }
    return token ;
  }

      //////////////////////////
      //
  void doinclude (string&)
    {
      LexerError ("doinclude() not implemented") ;
    }

      // Following functions should really by protected

  virtual int LexerInput (char*, int) ;
  virtual void LexerError (const char*) ;
  virtual void LexerInit () ;

protected:
  
  virtual const char* matchkeyword (const char*, int*) ;
  virtual const char* matchdirective (const char*, int*) ;

  void ScanStatement () ;
  void initkeys (struct Keylist *names, struct Keylist* [], struct Keylist* []) ;
  int is_data_statement () ;
  void checkBuffers (int) ;

private:

  ifstream in ;         // current input stream
  char* rawBuf ;        // characters read into here from istream
  char* scanBuf ;       // buffer for a complete pre-scanned source line
  char* cmntBuf ;       // comment buffer; keeps comments until next statement
  char* ip ;            // read-out pointer within rawBuf
  char* op ;            // end-of-statement pointer in scanBuf
  char* bp ;            // next character to be sent to lexer from scanBuf
  char* eos ;           // start of lookahead text in scanBuf
  char* cwp ;           // next free space in comment buffer
  char* crp ;           // next unread char in comment buffer
  char* eoc ;           // start of lookahear text in comment buffer
  char* end_of_scanBuf; // last usable location in scanBuf
  int nchars ;          // number of unread chars in rawBuf
  int buf_size ;        // size of both rawBuf and scanBuf
  int ip_position ;     // position of next input character on input line
  int holl_count ;      // number of hollerith characters yet to be scanned
  char* digit_start ;   // start pos in i/p buffer of last digit seq
  int dig_pos ;         // i/p line position of last digit seq
  int dig_copied ;      // no chars copied up to start of last digit seq
  int in_dig_seq ;      // == 1 if currently scanning a digit sequence
  int lines_scanned ;   // number of lines read by pre-scanner
  int next_stmt_line ;  // line number of next statement to be scanned
  int stmt_offset ;     // line number offset of next line to output
  int scan_state ;      // state variable for pre-scanner
  int old_state ;       // state pertaining on last statement line
  int is_blank ;        // set until current line has non-blank
  int is_continuation ; // set when current line is a continuation
  int pos_in_line ;     // position of character in line (fgg added on 2000/12/05)
};

#endif // _ftnLexer_h_
