/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// $Id: Types.cc,v 1.10 1999/06/04 09:07:09 npt Exp $

#include <string.h>
#include "Octave/Base/ParseContext.h"

// Static functions for Types.cc

static int computeIndentation (PTR_SCOPE) ;

// Member functions for class Datatype

void Datatype::set_fields (PTR_CNTXT p, type_tag v, int f)
{
  pc = p ;
  type_variant (v) ;
  length = 0 ;
  name = 0 ;
  type_flags = f ;
  node.fields.base_type = 0 ;
  node.fields.arg_types = 0 ;
  node.fields.ranges = 0 ;
  node.fields.dummy1 = 0 ;
  node.fields.dummy2 = 0 ;
  node.fields.dummy3 = 0 ;
  if (v == T_LIST)
    node.list = new TypeList ;
}

// TBD: copy constructor for type nodes is not yet complete

Datatype::Datatype (Datatype& d) : Tag (d.type_variant())
{
  name = d.name ;
  pc = d.pc ;
  length = d.length ;
  type_flags = d.type_flags ;
  node.fields.base_type = d.node.fields.base_type ;
  node.fields.arg_types = new TypeList (*(d.node.fields.arg_types)) ;
  node.fields.ranges = d.node.fields.ranges ; // TBD: make a copy
  node.fields.dummy1 = d.node.fields.dummy1 ;
  node.fields.dummy2 = d.node.fields.dummy2 ;
  node.fields.dummy3 = d.node.fields.dummy3 ;
}

Datatype::~Datatype ()
{
  type_tag tt = type_variant() ;
  if (tt == T_LIST && node.list)
    delete node.list ;
  else if (tt == T_FUNCTION && node.fields.arg_types)
    delete node.fields.arg_types ;
  else if ((tt == T_STRUCT || tt == T_UNION || tt == T_CLASS) &&
	   node.derived_class.obj_type)
    delete node.derived_class.obj_type ;
}

int operator== (Datatype &t1, Datatype &t2)
{
  if (t1.type_variant() == T_DESCRIPT)
    return *(t1.node.fields.base_type) == t2 ;

  if (t2.type_variant() == T_DESCRIPT)
    return t1 == *(t2.node.fields.base_type) ;

  type_tag tt1 = t1.type_variant() ;
  type_tag tt2 = t2.type_variant() ;

  if (tt1 != tt2)
    {
      // Types are dissimilar, UNLESS one is a T* and
      // the other is T[].

      if ((tt1 == T_POINTER && tt2 == T_ARRAY) ||
	  (tt2 == T_POINTER && tt1 == T_ARRAY))
	return *(t1.node.fields.base_type) == *(t2.node.fields.base_type) ;
      else
	return 0 ;
    }

  switch (t1.type_variant())
    {
    case T_DEFAULT:  // Scalar types only require equivalence 
    case T_INT:      // with their type flags (e.g. 'const'),
    case T_FLOAT:    // excluding the TYPEDEF, EXTERN, FINALISED, ERROR,
    case T_DOUBLE:   // STATIC, INLINE, and FRIEND flags.
    case T_CHAR:     //
    case T_BOOL:     //
    case T_STRING:   //
    case T_COMPLEX:  //
    case T_WCHART:   //
    case T_VOID:     //
      {
	const int incf = ~(GOT_EXTERN | GOT_TYPEDEF | FINALISED | GOT_ERROR
			   | GOT_STATIC | GOT_INLINE | GOT_FRIEND) ;
	return ((t1.type_flags & incf) == (t2.type_flags & incf)) ;
      }
    case T_ENUM:     // The type names must be the same
    case T_CLASS:    //
    case T_SUBRANGE: //
      return (t1.name == t2.name);
    case T_POINTER:  // Pointer type flags must match, and base types too.
      {
	const int incf = ~(GOT_EXTERN | GOT_TYPEDEF | GOT_STATIC | GOT_INLINE
			   | GOT_FRIEND | FINALISED | GOT_ERROR) ;
	if (t1.node.fields.base_type && t2.node.fields.base_type)
	  return (((t1.type_flags & incf) == (t2.type_flags & incf)) &&
		  (t1.node.fields.base_type == t2.node.fields.base_type ||
		  *(t1.node.fields.base_type) == *(t2.node.fields.base_type)
		   ));
	else
	  return 0 ;
      }
    case T_FUNCTION: 
      // Return types must match, and argument types must also match
      // if they are both defined (C only).
      // In C++ even empty argument lists must match

      if (t1.node.fields.base_type && t2.node.fields.base_type)
	{
	  int baseMatch = *(t1.node.fields.base_type) == *(t2.node.fields.base_type) ;
	  if (baseMatch)
	    {
	      TypeList* a1 = t1.node.fields.arg_types ;
	      TypeList* a2 = t2.node.fields.arg_types ;
	      if (!a1 || !a2)
		return 1 ;           // C only. In C++ we must do it differently!
	      else
		return *(a1) == *(a2) ;
	    }
	  else
	    return 0 ;
	}
      else
	return 0 ;
    case T_ARRAY:   // array types are equivalent if their base types are
                    // equivalent and their dimensions are equivalent.
      if (t1.node.fields.base_type && t2.node.fields.base_type)
	{
	  int baseMatch = *(t1.node.fields.base_type) == *(t2.node.fields.base_type) ;
	  if (baseMatch)
	    {
	      PTR_EXPR d1 = t1.node.fields.ranges ;
	      PTR_EXPR d2 = t2.node.fields.ranges ;
	      if (!d1 && !d2)
		return 1 ;
	      else {
		int v1, v2 ;
		if (d1->isIntConst(&v1) && d2->isIntConst(&v2))
		  return v1 == v2 ;
		else
		  return 0 ;
	      }
	    }
	  else
	    return 0 ;
	}
      else
	return 0 ;
    case T_UNION:   // unions and structs are equivalent if the types of
    case T_STRUCT:  // all their fields match exactly, or (trivially) if
                    // their type nodes have the same address, or if
                    // their symbolic names are identical.
      if (&t1 == &t2)
	return 1 ;
      else if (t1.name == t2.name)
	return 1 ;
      else if (t1.node.derived_class.obj_type && t2.node.derived_class.obj_type)
	if (*(t1.node.derived_class.obj_type) == *(t2.node.derived_class.obj_type))
	  {
	    t1.pc->warn ("compatible aggregate types have distinct definitions") ;
	    return 1 ;
	  }
	else
	  return 0 ;
      else
	return 0 ;
    case T_LIST:
      return *(t1.node.list) == *(t2.node.list) ;
    case T_RECORD:
    case T_ENUM_FIELD:
    case T_DESCRIPT:
    case T_DERIVED_CLASS:
    case T_DERIVED_TYPE:
    case T_GATE:
    case T_EVENT:
    case T_SEQUENCE:
      return 1 ;

    case T_UNKNOWN: // unknown types are always equivalent, just to
                    // cut down on unecessary error messages.
      return 1 ;

    default:
      return 0 ;
    }
}

int operator!= (Datatype &t1, Datatype &t2)
{
  return !(t1 == t2) ;
}

int operator== (TypeList &l1, TypeList &l2)
{
  TypeListIterator next1 (l1) ;
  TypeListIterator next2 (l2) ;
  PTR_TYPE t1 ;
  PTR_TYPE t2 ;
  
  while ( (t1 = next1(), t2 = next2(), t1 && t2 ) )
    if (*(t1) != *(t2))
      return 0 ;

  if (!t1 && !t2)
    return 1 ;
  else
    return 0 ;
}

int operator!= (TypeList &l1, TypeList &l2)
{
  return !(l1 == l2) ;
}

char* usual_conversions (PTR_TYPE* pt1, PTR_TYPE* pt2)
{
  if ((*pt1)->isLongDouble())
    return convert_type (pt2, T_DOUBLE, GOT_LONG) ;
  else if ((*pt2)->isLongDouble())
    return convert_type (pt1, T_DOUBLE, GOT_LONG) ;
  else if ((*pt1)->isDouble())
    return convert_type (pt2, T_DOUBLE, 0) ;
  else if ((*pt2)->isDouble())
    return convert_type (pt1, T_DOUBLE, 0) ;
  else if ((*pt1)->isFloat())
    return convert_type (pt2, T_FLOAT, 0) ;
  else if ((*pt2)->isFloat())
    return convert_type (pt1, T_FLOAT, 0) ;
  
  // apply integral promotions
  {
    char *emsg = integral_promotion (pt1) ;
    if (emsg)
      return emsg ;
    emsg = integral_promotion (pt2) ;
    if (emsg)
      return emsg ;
  }    
  
  if ((*pt1)->isUnsignedLong())
    return convert_type (pt2, T_INT, GOT_LONG+GOT_UNSIGNED) ;
  else if ((*pt2)->isUnsignedLong())
    return convert_type (pt1, T_INT, GOT_LONG+GOT_UNSIGNED) ;
  else if ((*pt1)->isLongInt() && (*pt2)->isUnsignedInt()
	   ||
	   (*pt2)->isLongInt() && (*pt1)->isUnsignedInt())
    {
      char* m1 = convert_type (pt1, T_INT, GOT_LONG+GOT_UNSIGNED) ;
      if (m1)
	return m1 ;
      else
	return convert_type (pt2, T_INT, GOT_LONG+GOT_UNSIGNED) ;
    }
  else
    {
      char* m1 = convert_type (pt1, T_INT, 0) ;
      if (m1)
	return m1 ;
      else
	return convert_type (pt2, T_INT, 0) ;
    }
}

char* convert_type (PTR_TYPE* t, type_tag tt, unsigned long tf)
{
  if ((*t)->isArithType() && IS_ARITH_TYPE(tt))
    {
      // If they are not identical create and link a new node
      if (tt != (*t)->type_variant() || tf != (*t)->type_flags)
	{
	  PTR_TYPE tmp = (*t)->pc->make_type (tt, tf) ;
	  tmp->node.fields.base_type = *t ;
	  *t = tmp ;
	}
      return 0 ;
    }
  else
    {
      PTR_TYPE tmp = (*t)->pc->make_type (tt, tf) ;
      return type_error ("convert", *t, tmp) ;
    }
}

char* pointer_conversions (PTR_EXPR e, PTR_TYPE t)
{
  int iv = 1 ;

  if (!e || !(e->type) || !t || (t->type_variant() != T_POINTER))
    return 0 ;

  type_tag tt = e->type->type_variant() ;

  switch (tt)
    {
    case T_POINTER:
      // OK if they point to same type, otherwise convert
      // e->type to void* if it points to an non-const/volatile object,
      // or if it points to a function.
      if (*(e->type) == *t)
	return 0 ;
      else
	{
	  PTR_TYPE bt = e->type->node.fields.base_type ;
	  if (bt->isAggregateType() && !(bt->type_flags & (GOT_CONST+GOT_VOLATILE))
	      || bt->isFuncType())
	    {
	      e->type = t->pc->make_type (T_POINTER, GOT_CONST) ;
	      e->type->node.fields.base_type = t->pc->global_void ;
	      return 0 ;
	    }
	  else
	    return type_error ("convert", e->type, t) ;
	}

    case T_ARRAY:
      // Convert `array-of-T' to `pointer-to-T' and try again.
      {
	PTR_TYPE tmp = t->pc->make_type (T_POINTER) ;
	tmp->node.fields.base_type = e->type->node.fields.base_type ;
	e->type = tmp ;
	return pointer_conversions (e, t) ;
      }

    default:
      // See if expression is constant integer zero...
      if (e->isIntConst(&iv))
	{
	  e->type = t->pc->make_type (T_POINTER, GOT_CONST) ;
	  e->type->node.fields.base_type = t->pc->global_void ;
	  if (iv)
	    {
	      char *emsg = "WARN:conversion from integer to pointer without cast" ;
	      char *buff = new char[strlen(emsg)+1] ;
	      strcpy(buff,emsg) ;
	      return buff ;
	    }
	  else
	    return 0 ;
	}
      else
	return type_error ("convert", e->type, t) ;
    }
}

char* integral_promotion (PTR_TYPE* t)
{
  if (!(*t))
    return 0 ;

  type_tag tt = (*t)->type_variant() ;

  switch (tt)
    {
    case T_INT:      return 0 ;
    case T_CHAR:     
    case T_SUBRANGE: return convert_type (t, T_INT, 0) ;
    default:         return type_error ("promote", *t, (*t)->pc->global_int) ;
    }
}

char* type_error (char* what, PTR_TYPE from, PTR_TYPE to)
{
  SourceListing *src = new SourceListing ;
  (*src) << "cannot " << what << " from " ;
  from->CcUnparseType (*src) ;
  (*src) << " to " ;
  to->CcUnparseType (*src) ;
  char *msg = src->Text() ;
  int msgLen = src->Size() ;
  char *buff = new char [msgLen+1] ;
  strcpy (buff, msg) ;
  delete src ;
  return buff ;
}

char* Datatype::sprintf ()
{
  SourceListing *src = new SourceListing ;
  CcUnparseType (*src) ;
  char *msg = src->Text() ;
  int msgLen = src->Size() ;
  char *buff = new char [msgLen+1] ;
  strcpy (buff, msg) ;
  delete src ;
  return buff ;
}

void Datatype::CcUnparseType (SourceListing& src, UnpCntl uc)
{
  if (!this) return ;

  TypeStack ts ;
  SymbolStack ss ;
  PTR_TYPE ds = this->dirdecl()->declspec () ;

  ds->unp_declspec (src, uc) ;
  this->unp_absdecl (src, ts, ss, 0) ;
  CcUnparseStack (src, ts, ss) ;
}

PTR_TYPE Datatype::declspec ()
{
  if (!this) return 0 ;
  switch (type_variant())
    {
    case T_FUNCTION: 
    case T_POINTER:
      return node.fields.base_type->declspec() ;
    case T_ARRAY: 
      return node.ar_decl.base_type->declspec() ;
    default:
      return this ;
    }
}

PTR_TYPE Datatype::dirdecl ()
{
  if (!this) return 0 ;
  switch (type_variant())
    {
    case T_POINTER: return node.fields.base_type->dirdecl() ;
    default:
      return this ;
    }
}

void Datatype::unp_absdecl (SourceListing& src, TypeStack &ts,
			    SymbolStack& ss, PTR_SYMB dsym)
{
  if (!this) return ;
  
  type_tag tt = type_variant() ;
  PTR_TYPE b = node.fields.base_type ;

  switch (tt)
    {
    case T_FUNCTION:
      b->unp_absdecl (src, ts, ss, dsym) ;
      ts.push (this) ;
      if (dsym && dsym->type == this)
	ss.push (dsym) ;
      else
	ss.push (0) ;
      break ;
    case T_POINTER:
      {
	if (b && (b->type_variant() == T_FUNCTION) || 
	    (b->type_variant() == T_ARRAY))
	  {
	    src << "(" ;
	    ts.push (this) ;
	  }
	CcUnparsePointer (src) ;
	b->unp_absdecl (src, ts, ss, dsym) ;
	break ;
      }
    case T_ARRAY :
      b->unp_absdecl (src, ts, ss, dsym) ;
      ts.push (this) ;
      break ;
    case T_DESCRIPT:
      if (dsym == name)
	b->unp_absdecl (src, ts, ss, dsym) ;
      else if (dsym)
	src << dsym->ident() ;
      break ;
    default:
      if (dsym)
	{
	  symbol_tag sst = dsym->symb_variant() ;
	  if (sst != ANON_FIELD_NAME)
	    src << dsym->ident() ;
	  if ((sst == FIELD_NAME || sst == ANON_FIELD_NAME) &&
	      dsym->node.field.restricted_bit)
	    src << " : " ;
	  dsym->node.field.restricted_bit->CcUnparseExpression (src, 0) ;
	}
      break ;
    }
}

void Datatype::CcUnparsePointer (SourceListing& src)
{
  src << "* " ;
  if (type_flags & GOT_CONST) src << "const " ;
  if (type_flags & GOT_VOLATILE) src << " volatile " ;
}

void Datatype::unp_declspec (SourceListing& src, UnpCntl uc)
{
  if (!this)
    return;

  type_tag tt = type_variant();

  if (type_flags & GOT_TYPEDEF) src << "typedef " ;
  if (type_flags & GOT_EXTERN) src << "extern " ;
  if (type_flags & GOT_STATIC) src << "static " ;
  if (type_flags & GOT_AUTO) src << "auto " ;
  if (type_flags & GOT_REGISTER) src << "register " ;
  if (type_flags & GOT_INLINE) src << "inline " ;
  if (type_flags & GOT_CONST) src << "const " ;
  if (type_flags & GOT_VOLATILE) src << "volatile " ;
  if (type_flags & GOT_SIGNED) src << "signed " ;
  if (type_flags & GOT_UNSIGNED) src << "unsigned " ;
  if (type_flags & GOT_LONG) src << "long " ;
  if (type_flags & GOT_SHORT) src << "short " ;

  if (name && tt == T_DESCRIPT)
    {
      src << name->ident() ;
      return ;
    }

  switch (tt)
    {
    case T_INT: src << "int" ; break ;
    case T_FLOAT: src << "float" ; break ;
    case T_DOUBLE: src << "double" ; break ;
    case T_CHAR: src << "char" ; break ;
    case T_VOID: src << "void" ; break ;

    case T_STRUCT:
    case T_UNION:
      {
	if (tt == T_STRUCT)
	  src << "struct " ;
	else
	  src << "union " ;

	/* Unparse the full structure definition if it is defined and
	 * the callee requested a full definition (via the 'uc' arg)
	 * OR if the struct name is anonymous.
	 */

	if ((node.derived_class.num_fields >= 0 && uc == UNP_VERBOSE)
	    || (name && (name->attr & ATT_ANON)))
	  {
	    if (name && !(name->attr & ATT_ANON))
	      src << name->ident() << " { " ;
	    else
	      src << "{ " ;

	    if (node.derived_class.obj_type)
	      if (node.derived_class.decls)
		{
		  // Unparsing a concrete type, using the StmtList*
		  // field 'node.derived_class.decls'

		  src << '\n' ;
		  src.appendBuffer () ;
		  PTR_SCOPE owner = node.derived_class.members ;
		  PTR_STMT  dstmt = node.derived_class.original_class ;
		  int tb = computeIndentation (owner) - 1 ;
		  StmtListIterator next_decl (*node.derived_class.decls) ;
		  PTR_STMT stmt ;
		  while (stmt = next_decl())
		    {
		      src.CcStartLine (tb, dstmt, 0) ;
		      if (stmt->stmt_variant() == VAR_DECL)
			{
			  stmt->node.fields.e1->CcUnparseExpression (src, 0) ;
			  if (stmt->node.fields.e2)
			    {
			      src << " " ;
			      stmt->node.fields.e2->CcUnparseExpression (src, 0) ;
			    }
			}
		      else
			src << "/* UNPARSE ERROR : expecting declaration */" ;

		      src << " ;" ;
		      src << '\n' ;
		      src.appendBuffer () ;
		      src.CcStartLine (tb-1, dstmt, 0) ;
		    }
		}
	      else
		{
		  // Unparsing an abstract type declaration or a concrete
		  // type in an abstract way.
		  //
		  TypeListIterator nextType (*node.derived_class.obj_type) ;
		  PTR_TYPE ft = nextType();
		  while (ft)
		    {
		      ft->CcUnparseType (src, uc) ;
		      ft = nextType () ;
		      if (ft)
			src << ", " ;
		    }
		}
	    src << "}" ;
	  }
	else
	  {
	    if (name)
	      src << name->ident() ;
	  }
	break ;
      }

    case T_UNKNOWN:   // an ellipsis within a parameter list
      src << "..." ;
      break ;

    default:
      {
	TypeStack ts ;
	SymbolStack ss ;
	unp_absdecl (src, ts, ss, 0) ;
      }
      break ;
    }
}

void Datatype::unp_dirdecl (SourceListing& src, TypeStack &ts, SymbolStack& ss)
{
  if (!this) return;
  type_tag tt = type_variant();
  
  switch (tt)
    {
    case T_FUNCTION:
      node.fields.base_type->unp_dirdecl (src, ts, ss) ;
      ts.push (this) ;
      break ;
    case T_ARRAY:
      node.fields.base_type->unp_dirdecl (src, ts, ss) ;
      ts.push (this) ;
      break ;
    case T_POINTER:
      unp_absdecl (src, ts, ss, 0) ;
      break ;
    default:
      break ;
    }
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                static int computeIndentation                        |
 |                                                                     |
 +---------------------------------------------------------------------*/

static int computeIndentation (PTR_SCOPE sc)
{
  int count = 0;
  
  while (sc)
  {
    count++;
    sc = sc->enclosing_scope() ;
  }

  return count ;
}
