#include "Octave/Base/ParseTree.h"
#include "Octave/Fortran/FortranParser.h"
#include "Octave/Base/Unparse.h"

#define YYDEBUG
#ifdef YYDEBUG
extern int yydebug ;
#endif // YYDEBUG

#define NO_LISTING	0
#define SOURCE_LISTING	1
#define DEBUG_LISTING	2

#define QUIET		0
#define VERBOSE		1

void subroutineListing (PTR_STMT proc, dialect dlt, int listFlag)
{
  SourceListing list ;
  list.unparse (proc, ANSII_F77) ;
  SourceListingIterator next_line (list) ;
  SourceLine *line ;
  
  while (line = next_line() )
    if (listFlag == DEBUG_LISTING)
      printf ("%5d 0x%08x %s", line->lineNumber, (unsigned)(line->stmt), (*line)()) ;
    else
      printf ("%s", (*line)()) ;
}

void listing (FortranParser *fp, int listFlag)
{
  int numErrors = 0 ;
  slist_iterator<char*, string> next_error (fp->errors()) ;
  string *msg ;
  
  while (msg = next_error() )
  {
    ++numErrors ;
    fprintf (stderr, "%s", (*msg)()) ;
  }
  
  if (listFlag && (numErrors == 0)) {
    // PTR_STMT root = fp->fileRoot() ;
    
    StmtListIterator next_routine (fp->fileRoot()->then_part()) ;
    PTR_STMT subprog ;

    while ((subprog = next_routine()))
      subroutineListing (subprog, ANSII_F77, listFlag) ;
  }
}

void printRoutineNames (FortranParser *ftn)
{
  StmtListIterator next_routine (ftn->fileRoot()->then_part()) ;
  PTR_STMT stmt ;
  
  while ((stmt = next_routine()))
  {
    switch (stmt->stmt_variant())
    {
    case PROG_HEDR:
    case PROC_HEDR:
    case FUNC_HEDR:
      printf ("\t%s\n", stmt->symbol()->ident()) ;
      break ;
    default:
      break ;
    }
  }
}

void forceEndDo (FortranParser *ftn)
{
  StmtListIterator next_routine (ftn->fileRoot()->then_part()) ;
  PTR_STMT stmt ;
  
  while ((stmt = next_routine()))
  {
    switch (stmt->stmt_variant())
    {
    case PROG_HEDR:
    case PROC_HEDR:
    case FUNC_HEDR:
      printf ("Changing DO/CONTINUE to DO/ENDDO in '%s'\n",
	      stmt->symbol()->ident()) ;
      stmt->enforce_enddo() ;
      break ;
    default:
      break ;
    }
  }
}

void usage (char *option)
{
  fprintf (stderr, "test: illegal option '%s'\n", option) ;
  fprintf (stderr, "usage: test [-l] [-d] [-s] [-v] [-q] source.f\n") ;
  exit (EXIT_FAILURE) ;
}

int main (int argc, char *argv[])
{
  int lflag = NO_LISTING ;
  int vflag = QUIET ;
  int rflag = 0 ;
  int doflag = 0 ;
  string *fname ;

  yydebug = 0 ;

  if (argc < 2)
    exit (0) ;

  for (int p = 1; p < argc; p++)
  {
    if (!strcmp (argv[p], "-l"))
      lflag = SOURCE_LISTING ;
    else if (!strcmp (argv[p], "-d"))
      lflag = DEBUG_LISTING ;
    else if (!strcmp (argv[p], "-s"))
      rflag = 1 ;
    else if (!strcmp (argv[p], "-v"))
      vflag = VERBOSE ;
    else if (!strcmp (argv[p], "-dy"))
      yydebug = 1 ;
    else if (!strcmp (argv[p], "-q"))
      vflag = QUIET ;
    else if (!strcmp (argv[p], "-do"))
      doflag = 1 ;
    else if (*(argv[p]) == '-')
      usage (argv[p]) ;
    else
      fname = new string (argv[p]) ;
  }

  if (vflag)
    fprintf (stderr, "Creating FortranParser for file '%s'\n", (*fname)()) ;

  FortranParser *ftn = new FortranParser (*fname) ;

  if (vflag)
    fprintf (stderr, "Parser creation complete\n") ;

  int parseErrors = ftn->parse () ;
  
  if (vflag)
    fprintf (stderr, "Parse completed with %d errors\n", parseErrors) ;

  if (doflag && !parseErrors)
    forceEndDo (ftn) ;

  if (rflag && !parseErrors)
    printRoutineNames (ftn) ;

  listing (ftn, lflag) ;

  if (vflag)
    fprintf (stderr, "Destroying parse context\n") ;

  delete ftn ;

  if (vflag)
    fprintf (stderr, "Done.\n") ;
}
