/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// Header for class definition of TransHist objects - each such object
// records a single transformation to the source.

#if ! defined _transHistory_h_
#define _transHistory_h_

#include "Sets.h"
#include "Strings.h"
#include "Globals.h"
#include "ParseTree.h"

/////////////////////
// 
//  Tag values identifying categories of transformation
//
typedef enum trans_tag {
  HOLL_TRANS,           // Hollerith => char_string (lexical)
  ALLOC_PAREN,          // allocatable parentheses  (syntactic)
  PRINT_NML,            // print nml => write       (syntactic)
  READ_NML,             // read nml => read         (syntactic)
  ACCEPT_IO,            // accept f => read         (syntactic)
  ACCEPT_NONE,          // accept * => read         (syntactic)
  ACCEPT_NML,           // accept nml => read       (syntactic)
  TYPE_IO,              // type f => read           (syntactic)
  TYPE_NONE,            // type * => read           (syntactic)
  TYPE_NML,             // type nml => read         (syntactic)
  ENCODE_NCTL,          // encode (c,f,b)           (syntactic)
  ENCODE_WCTL,          // encode (c,f,b,ctlist)    (syntactic)
  DECODE_NCTL,          // decode (c,f,b)           (syntactic)
  DECODE_WCTL,          // decode (c,f,b,ctlist)    (syntactic)
  FIND_STMT,            // find (ctlist)            (syntactic)
  NULL_TRANS            // identity transformation
} trans_tag;

// CLASS GROUP: transformation records

/*--------------------------------------------------------------------

CLASS
    TransHist

    A class for recording source-level transformations

OVERVIEW

    Each source-level transformation is recorded within a single
    unique TransHist object. Each Statement, Expression, Symbol or
    Datatype that is transformed maintains a TransHist* pointer to
    the TransHist object which describes that transformation.
    The full set of TransHist objects is collected together in a
    Universe of TransHist objects, which is part of the ParseContext.

DESCRIPTION

    A full description will be provided later.

---------------------------------------------------------------------*/

class TransHist : public Enumerable {
  
public:
  
    ////////////////////////////////////
    //
    // Constructor.
    //
    //
  TransHist (int i) : Enumerable(i) { }

  int variant () { return 0 ; }

    ////////////////////////////////////
    //
    // After construction within the note_transform() function
    // within the ParseContext class, the fields of the
    // TransHist object can be set with this function
    //
    // [in] trans_tag     transformation category           
    // [in] Statement*    transformed statement
    //
  void set_fields (trans_tag t, PTR_STMT s)
    { ttag = t ; changed.stmt = s ; }

    ////////////////////////////////////
    //
    // After construction within the note_transform() function
    // within the ParseContext class, the fields of the
    // TransHist object can be set with this function
    //
    // [in] trans_tag     transformation category           
    // [in] Symbol*       transformed symbol
    //
  void set_fields (trans_tag t, PTR_SYMB s)
    { ttag = t ; changed.symb = s ; }

    ////////////////////////////////////
    //
    // After construction within the note_transform() function
    // within the ParseContext class, the fields of the
    // TransHist object can be set with this function
    //
    // [in] trans_tag     transformation category           
    // [in] Expression*   transformed expression
    //
  void set_fields (trans_tag t, PTR_EXPR s)
    { ttag = t ; changed.expr = s ; }

    ////////////////////////////////////
    //
    // After construction within the note_transform() function
    // within the ParseContext class, the fields of the
    // TransHist object can be set with this function
    //
    // [in] trans_tag     transformation category           
    // [in] Datatype*     transformed type node
    //
  void set_fields (trans_tag t, PTR_TYPE s)
    { ttag = t ; changed.type = s ; }

    ////////////////////////////////////
    //
    // Extracts the Statement implicated by this transformation
    //
    //  [out] Statement*  the statement node
    //
  PTR_STMT stmt_node ()     { return changed.stmt ; }

    ////////////////////////////////////
    //
    // Extracts the Expression implicated by this transformation
    //
    //  [out] Expression*  the expression node
    //
  PTR_EXPR expr_node ()     { return changed.expr ; }

    ////////////////////////////////////
    //
    // Extracts the Symbol implicated by this transformation
    //
    //  [out] Symbol*  the symbol node
    //
  PTR_SYMB symb_node ()     { return changed.symb ; }

    ////////////////////////////////////
    //
    // Extracts the Datatype implicated by this transformation
    //
    //  [out] Datatype*  the type node
    //
  PTR_TYPE type_node ()     { return changed.type ; }

private:

  int id ;
  trans_tag ttag ;
  union {
    PTR_STMT stmt ;
    PTR_EXPR expr ;
    PTR_SYMB symb ;
    PTR_TYPE type ;
  } changed ;
};

#endif // _tranHistory_h_
